package evolvekernel;

import evolvekernel.func.FastPow;
import libsvm.*;
/**
 * Utility class to provide various Mathematical computations liek
 * first order Bessels, power computations, Spline values of orders
 * factorials and binomial
 * @author udaykamath
 *
 */
public class Functions {
	private static final double SQRTPI = 2.50662827463100050242E0;

	public static double dotProduct(svm_node[] x, svm_node[] y) {
		double sum = 0;
		int xlen = x.length;
		int ylen = y.length;
		int i = 0;
		int j = 0;
		while (i < xlen && j < ylen) {
			if (x[i].index == y[j].index)
				sum += x[i++].value * y[j++].value;
			else {
				if (x[i].index > y[j].index)
					++j;
				else
					++i;
			}
		}
		return sum;
	}

	public static void copySVMNodeArray(svm_node[] x1, double[] x2) {
		// x2 = new double[x1.length];
		for (int i = 0; i < x1.length; i++) {
			x2[i] = x1[i].value;
		}
	}

	public static svm_node[] copySVMNode(svm_node[] x) {
		svm_node[] y = (svm_node[]) x.clone();
		return y;
	}

	// / <summary>
	// / Returns the Bessel function of order 0 of the specified number.
	// / </summary>
	public static double besselJ0(double x) {
		double ax;

		if ((ax = Math.abs(x)) < 8.0) {
			double y = x * x;
			double ans1 = 57568490574.0
					+ y
					* (-13362590354.0 + y
							* (651619640.7 + y
									* (-11214424.18 + y
											* (77392.33017 + y * (-184.9052456)))));
			double ans2 = 57568490411.0
					+ y
					* (1029532985.0 + y
							* (9494680.718 + y
									* (59272.64853 + y
											* (267.8532712 + y * 1.0))));

			return ans1 / ans2;

		} else {
			double z = 8.0 / ax;
			double y = z * z;
			double xx = ax - 0.785398164;
			double ans1 = 1.0
					+ y
					* (-0.1098628627e-2 + y
							* (0.2734510407e-4 + y
									* (-0.2073370639e-5 + y * 0.2093887211e-6)));
			double ans2 = -0.1562499995e-1
					+ y
					* (0.1430488765e-3 + y
							* (-0.6911147651e-5 + y
									* (0.7621095161e-6 - y * 0.934935152e-7)));

			return Math.sqrt(0.636619772 / ax)
					* (Math.cos(xx) * ans1 - z * Math.sin(xx) * ans2);
		}
	}

	// / <summary>
	// / Returns the Bessel function of order 1 of the specified number.
	// / </summary>
	public static double besselJ(double x) {
		double ax;
		double y;
		double ans1, ans2;

		if ((ax = Math.abs(x)) < 8.0) {
			y = x * x;
			ans1 = x
					* (72362614232.0 + y
							* (-7895059235.0 + y
									* (242396853.1 + y
											* (-2972611.439 + y
													* (15704.48260 + y
															* (-30.16036606))))));
			ans2 = 144725228442.0
					+ y
					* (2300535178.0 + y
							* (18583304.74 + y
									* (99447.43394 + y
											* (376.9991397 + y * 1.0))));
			return ans1 / ans2;
		} else {
			double z = 8.0 / ax;
			double xx = ax - 2.356194491;
			y = z * z;

			ans1 = 1.0
					+ y
					* (0.183105e-2 + y
							* (-0.3516396496e-4 + y
									* (0.2457520174e-5 + y * (-0.240337019e-6))));
			ans2 = 0.04687499995
					+ y
					* (-0.2002690873e-3 + y
							* (0.8449199096e-5 + y
									* (-0.88228987e-6 + y * 0.105787412e-6)));
			double ans = Math.sqrt(0.636619772 / ax)
					* (Math.cos(xx) * ans1 - z * Math.sin(xx) * ans2);
			if (x < 0.0)
				ans = -ans;
			return ans;
		}
	}

	// / <summary>
	// / Returns the Bessel function of order n of the specified number.
	// / </summary>
	public static double besselJ(int n, double x) {
		int j, m;
		double ax, bj, bjm, bjp, sum, tox, ans;
		boolean jsum;

		double ACC = 40.0;
		double BIGNO = 1.0e+10;
		double BIGNI = 1.0e-10;

		if (n == 0)
			return besselJ0(x);
		if (n == 1)
			return besselJ(x);

		ax = Math.abs(x);
		if (ax == 0.0)
			return 0.0;
		else if (ax > (double) n) {
			tox = 2.0 / ax;
			bjm = besselJ0(ax);
			bj = besselJ(ax);
			for (j = 1; j < n; j++) {
				bjp = j * tox * bj - bjm;
				bjm = bj;
				bj = bjp;
			}
			ans = bj;
		} else {
			tox = 2.0 / ax;
			m = 2 * ((n + (int) Math.sqrt(ACC * n)) / 2);
			jsum = false;
			bjp = ans = sum = 0.0;
			bj = 1.0;
			for (j = m; j > 0; j--) {
				bjm = j * tox * bj - bjp;
				bjp = bj;
				bj = bjm;
				if (Math.abs(bj) > BIGNO) {
					bj *= BIGNI;
					bjp *= BIGNI;
					ans *= BIGNI;
					sum *= BIGNI;
				}
				if (jsum)
					sum += bj;
				jsum = !jsum;
				if (j == n)
					ans = bjp;
			}
			sum = 2.0 * sum - bj;
			ans /= sum;
		}
		return x < 0.0 && n % 2 == 1 ? -ans : ans;
	}

	// / <summary>
	// / Returns the Bessel function of the second kind, of order 0 of the
	// specified number.
	// / </summary>
	public static double BesselY0(double x) {
		if (x < 8.0) {
			double y = x * x;

			double ans1 = -2957821389.0
					+ y
					* (7062834065.0 + y
							* (-512359803.6 + y
									* (10879881.29 + y
											* (-86327.92757 + y * 228.4622733))));
			double ans2 = 40076544269.0
					+ y
					* (745249964.8 + y
							* (7189466.438 + y
									* (47447.26470 + y
											* (226.1030244 + y * 1.0))));

			return (ans1 / ans2) + 0.636619772 * besselJ0(x) * Math.log(x);
		} else {
			double z = 8.0 / x;
			double y = z * z;
			double xx = x - 0.785398164;

			double ans1 = 1.0
					+ y
					* (-0.1098628627e-2 + y
							* (0.2734510407e-4 + y
									* (-0.2073370639e-5 + y * 0.2093887211e-6)));
			double ans2 = -0.1562499995e-1
					+ y
					* (0.1430488765e-3 + y
							* (-0.6911147651e-5 + y
									* (0.7621095161e-6 + y * (-0.934945152e-7))));
			return Math.sqrt(0.636619772 / x)
					* (Math.sin(xx) * ans1 + z * Math.cos(xx) * ans2);
		}
	}

	// /
	// / Computes the Basic Spline of order n
	// / </summary>
	public static double bSpline(int n, double x) {
		// ftp://ftp.esat.kuleuven.ac.be/pub/SISTA/hamers/PhD_bhamers.pdf
		// http://sepwww.stanford.edu/public/docs/sep105/sergey2/paper_html/node5.html

		double a = 1.0 / factorial(n);
		double c;

		boolean positive = true;
		for (int k = 0; k <= n + 1; k++) {
			c = binomial(n + 1, k) * truncPower(x + (n + 1.0) / 2.0 - k, n);
			a += positive ? c : -c;
			positive = !positive;
		}

		return a;
	}

	public static double truncPower(double value, double degree) {
		double x = FastPow.pow(value, degree);
		return (x > 0) ? x : 0.0;
	}

	// /
	// / Returns the factorial of the specified value.
	// / </summary>
	public static double factorial(double value) {
		double d = Math.abs(value);

		// Check if the number is an integer
		if (Math.floor(d) == d) {
			// Calculate factorial iteractively
			return (double) factorial((int) value);
		} else {
			// Return Gamma approximation
			return gamma(value + 1.0);
		}
	}

	public static int factorial(int value) {
		int i = Math.abs(value);
		int fac = 1;

		while (i > 1) {
			fac *= i--;
		}

		// Return factorial with original value sign
		return (value < 0) ? -fac : fac;
	}

	public static double gamma(double x) {
		double[] P = { 1.60119522476751861407E-4, 1.19135147006586384913E-3,
				1.04213797561761569935E-2, 4.76367800457137231464E-2,
				2.07448227648435975150E-1, 4.94214826801497100753E-1,
				9.99999999999999996796E-1 };
		double[] Q = { -2.31581873324120129819E-5, 5.39605580493303397842E-4,
				-4.45641913851797240494E-3, 1.18139785222060435552E-2,
				3.58236398605498653373E-2, -2.34591795718243348568E-1,
				7.14304917030273074085E-2, 1.00000000000000000320E0 };

		double p, z;

		double q = Math.abs(x);

		if (q > 33.0) {
			if (x < 0.0) {
				p = Math.floor(q);

				if (p == q)
					throw new ArithmeticException();

				z = q - p;
				if (z > 0.5) {
					p += 1.0;
					z = q - p;
				}
				z = q * Math.sin(Math.PI * z);

				if (z == 0.0)
					throw new ArithmeticException();

				z = Math.abs(z);
				z = Math.PI / (z * stirling(q));

				return -z;
			} else {
				return stirling(x);
			}
		}

		z = 1.0;
		while (x >= 3.0) {
			x -= 1.0;
			z *= x;
		}

		while (x < 0.0) {
			if (x == 0.0) {
				throw new ArithmeticException();
			} else if (x > -1.0E-9) {
				return (z / ((1.0 + 0.5772156649015329 * x) * x));
			}
			z /= x;
			x += 1.0;
		}

		while (x < 2.0) {
			if (x == 0.0) {
				throw new ArithmeticException();
			} else if (x < 1.0E-9) {
				return (z / ((1.0 + 0.5772156649015329 * x) * x));
			}

			z /= x;
			x += 1.0;
		}

		if ((x == 2.0) || (x == 3.0))
			return z;

		x -= 2.0;
		p = polevl(x, P, 6);
		q = polevl(x, Q, 7);
		return z * p / q;

	}

	public static double stirling(double x) {
		double[] STIR = { 7.87311395793093628397E-4,
				-2.29549961613378126380E-4, -2.68132617805781232825E-3,
				3.47222221605458667310E-3, 8.33333333333482257126E-2, };
		double MAXSTIR = 143.01608;

		double w = 1.0 / x;
		double y = Math.exp(x);

		w = 1.0 + w * polevl(w, STIR, 4);

		if (x > MAXSTIR) {
			double v = FastPow.pow(x, 0.5 * x - 0.25);
			y = v * (v / y);
		} else {
			y = FastPow.pow(x, x - 0.5) / y;
		}
		y = SQRTPI * y * w;
		return y;
	}

	public static double polevl(double x, double[] coef, int N) {
		double ans;

		ans = coef[0];

		for (int i = 1; i <= N; i++) {
			ans = ans * x + coef[i];
		}

		return ans;
	}

	// /
	// / Computes the Binomial Coefficients C(n,k).
	// /
	public static double binomial(int n, int k) {
		double[] b = new double[n + 1];
		b[0] = 1;

		for (int i = 1; i <= n; ++i) {
			b[i] = 1;
			for (int j = i - 1; j > 0; --j)
				b[j] += b[j - 1];
		}

		return b[k];
	}

}