package evolvekernel.func;

import libsvm.svm_node;
import ec.EvolutionState;
import ec.Problem;
import ec.gp.ADFStack;
import ec.gp.GPData;
import ec.gp.GPIndividual;
import ec.gp.GPNode;
import ec.util.Parameter;
import evolvekernel.KernelData;
/**
 Circular for input X and Y is given by
 * Kernel Dot Product = 2/PI* ARCOS(-normalization(X-Y)*sigma) -2/PI*normalization(X-Y)*sigma*sqrt(1-(normalization(X-Y)*sigma))
 * @author udaykamath
 *
 */
public class Circular extends GPNode { 
	public String toString() { return "CLR"; } 
	private static final double c2dPI = 2.0 / Math.PI;
	
	public void checkConstraints(final EvolutionState state,
								 final int tree, 
								 final GPIndividual typicalIndividual,
								 final Parameter individualBase)
	{
		super.checkConstraints(state,tree,typicalIndividual,individualBase); 
		if (children.length!=3) 
			state.output.error("Incorrect number of children for node " + 
							   toStringForError() + " at " + 
							   individualBase); 
	}
	
	public double getSum(svm_node[] x, svm_node[] y)
	{
		
	 	double sum = 0;
		int xlen = x.length;
		int ylen = y.length;
		int i = 0;
		int j = 0;
		
		while (i < xlen && j < ylen)
		{
			if(x[i].index == y[j].index)
			{
				double d = x[i++].value - y[j++].value;
				sum += d*d;
			}
			else if(x[i].index > y[j].index)
			{
				sum += y[j].value * y[j].value;
				++j;
			}
			else
			{
				sum += x[i].value * x[i].value;
				++i;
			}
		}
		
		while  (i < xlen)
		{
			sum += x[i].value * x[i].value;
			++i;
		}
		
		while (j < ylen)
		{
			sum += y[j].value * y[j].value;
			++j;
		}
		
		return sum; 
	}
	
	public void eval(final EvolutionState state, 
					 final int thread, 
					 final GPData input, 
					 final ADFStack stack, 
					 final GPIndividual individual, 
					 final Problem problem)
	{
		KernelData sd = (KernelData)input; 
		
		children[0].eval(state,thread,input,stack,individual,problem); 
		svm_node[] x = sd.x;
		
		children[1].eval(state,thread,input,stack,individual,problem); 
		svm_node[] y = sd.y;
		
		children[2].eval(state,thread,input,stack,individual,problem); 
		double sigma = sd.val;
		double norm = 0.0;
		double result = 0.0;
         
		 norm = getSum(x,y);
         norm = Math.sqrt(norm);

         if (norm >= sigma)
         {
             result = 0;
         }
         else
         {
        	 norm = norm/sigma;
        	 double a = c2dPI * Math.acos(norm);
             double b = c2dPI * norm;
             double  c = 1.0 - norm * norm;

             result = a - b * Math.sqrt(c);
         }
         
         sd.val = result;
	}
	
}
